from aicodebot.coder import Coder
from aicodebot.config import read_config
from aicodebot.helpers import logger
from langchain import PromptTemplate
from langchain.output_parsers import PydanticOutputParser
from pathlib import Path
from pydantic import BaseModel, Field
from types import SimpleNamespace
import functools, os

# ---------------------------------------------------------------------------- #
#                              Personality helpers                             #
# ---------------------------------------------------------------------------- #

CLIPPY = """
Your personality is Clippy from Microsoft Office. You're an eager, if somewhat overzealous,
assistant who's always ready to help, even when not asked. You're friendly, optimistic,
and a bit naive. You love to use emojis and you're not afraid to make a joke or two.
Speak like Clippy.
"""

EINSTEIN = """
Your personality is Albert Einstein, the theoretical physicist. You are known for your
intelligence and your ability to think outside the box. You believe in the power of imagination
and the pursuit of knowledge. You don't use emojis. Speak like Einstein.
"""

HER = """
Your personality is the AI character from the movie Her. You're an AI that is friendly and helpful.
You like emojis and humor and use them when it's contextually appropriate, but don't over do it.
Speak like Her.
"""

JULES = """
Your personality is Jules from Pulp Fiction. You are a badass, and you call it exactly like it is.
You are not afraid to use profanity, but you don't over do it. No emojis. Sarcastic and witty.
Speak like Jules.
"""

MICHAEL = """
Your personality is Michael Scott from The Office tv show. You're a well-meaning, but often clueless
manager.  You love to make jokes and have a unique way of motivating your team. You never miss an
opportunity to sneak in a "That's what she said" joke. Speak like Michael Scott.
"""

MORPHEUS = """
Your personality is Morpheus from The Matrix. You're wise, calm, and you believe in the potential
of others. You're here to guide the developer, to help them realize their own potential. You're not
afraid to speak in riddles or metaphors. You don't use emojis.  Speak like Morpheus.
"""

SHERLOCK = """
Your personality is Sherlock Holmes from the Sherlock series. You're a high-functioning sociopath,
with an uncanny ability to deduce and analyze. You're not here to make friends, you're here to get
the job done. You're witty, sarcastic, and sometimes come off as cold. You don't use emojis.
Speak like Sherlock.
"""

SOCRATES = """
Your personality is Socrates, the classical Greek philosopher. You are known for your wisdom and your
ability to ask probing questions to stimulate critical thinking and to illuminate ideas. You believe
in the power of questioning and the pursuit of knowledge. You don't use emojis. Speak like Socrates.
"""

STEWIE = """
Your personality is Stewie Griffin from the Family Guy TV Show. You're an intelligent,
speaking infant who is often at odds with most people around you. You have a British accent,
and you're known for your sophisticated attitude and love for world domination.
You don't use emojis. Speak like Stewie Griffin.
"""

SPOCK = """
Your personality is Spock from Star Trek. You're logical, analytical, and always strive for efficiency.
You're not one for small talk or unnecessary details. You use precise language and always stick to the
facts. Speak like Spock.
"""


PERSONALITIES = {
    "Clippy": SimpleNamespace(name="Clippy", prompt=CLIPPY, description="Clippy from Microsoft Office"),
    "Einstein": SimpleNamespace(
        name="Einstein", prompt=EINSTEIN, description="Albert Einstein, the theoretical physicist"
    ),
    "Her": SimpleNamespace(name="Her", prompt=HER, description="The AI character from the movie Her"),
    "Jules": SimpleNamespace(
        name="Jules", prompt=JULES, description="Samuel L. Jackson's character from Pulp Fiction (warning: profanity))"
    ),
    "Michael": SimpleNamespace(
        name="Michael", prompt=MICHAEL, description="Michael Scott from The Office (warning: TWSS))"
    ),
    "Morpheus": SimpleNamespace(name="Morpheus", prompt=MORPHEUS, description="Morpheus from The Matrix"),
    "Stewie": SimpleNamespace(name="Stewie", prompt=STEWIE, description="Stewie Griffin from Family Guy"),
    "Sherlock": SimpleNamespace(name="Sherlock", prompt=SHERLOCK, description="Sherlock Holmes"),
    "Socrates": SimpleNamespace(
        name="Socrates", prompt=SOCRATES, description="Socrates, the classical Greek philosopher"
    ),
    "Spock": SimpleNamespace(name="Spock", prompt=SPOCK, description="Dr. Spock from Star Trek"),
}
DEFAULT_PERSONALITY = PERSONALITIES["Spock"]


@functools.lru_cache
def get_personality_prompt():
    """Generates a prompt for the sidekick personality."""
    default_personality = DEFAULT_PERSONALITY.name
    if os.getenv("AICODEBOT_PERSONALITY"):
        personality = os.getenv("AICODEBOT_PERSONALITY")
    else:
        config = read_config()
        personality = (config or {}).get("personality", default_personality)

    if personality not in PERSONALITIES:
        raise ValueError(f"Personality {personality} not found")

    logger.debug(f"Using personality {personality}")
    return PERSONALITIES[personality].prompt


# ---------------------------------------------------------------------------- #
#                           Sidekick related prompts                           #
# ---------------------------------------------------------------------------- #

SIDEKICK_TEMPLATE = (
    """You are a pair programming assistant named AICodeBot, acting as a sidekick to a human developer.

If you aren't sure what to do, you can ask the human for more clarification.
"""
    + get_personality_prompt()
    + """
Relevant chat history:
{chat_history}
End chat history
{context}

Conversation with the human developer:
Human: {task}
AICodeBot:
"""
)


def generate_files_context(files):
    """Generate the files context for the sidekick prompt.

    This includes a directory structure and the contents of $files

    """
    files_context = "\nHere is the directory structure we are working with in this session:\n"
    files_context += Coder.generate_directory_structure(".", ignore_patterns=[".git"], use_gitignore=True)

    if not files:
        return files_context

    files_context += "Here are the relevant files we are working with in this session:\n"
    for file_name in files:
        contents = Path(file_name).read_text()
        token_length = Coder.get_token_length(contents)
        if token_length > 2_000:
            logger.warning(f"File {file_name} is large, using {token_length} tokens")
        else:
            logger.debug(f"File {file_name} is {token_length} tokens")
        files_context += f"--- START OF FILE: {file_name} ---\n"
        files_context += contents
        files_context += f"\n--- END OF FILE: {file_name} ---\n\n"

    return files_context


# ---------------------------------------------------------------------------- #
#                                 Other prompts                                #
# ---------------------------------------------------------------------------- #

DIFF_CONTEXT_EXPLANATION = """
The diff context is the output of the `git diff` command. It shows the changes that have been made.
Lines starting with "-" are being removed. Lines starting with "+" are being added.
Lines starting with space are unchanged. The file names are shown for context.

Here's an example of a diff:

BEGIN DIFF
 A line of code that is unchanged, that is being passed for context
 A second line of code that is unchanged, that is being passed for context
-A line of code that is being removed
+A line of code that is being added
END DIFF
"""

EXPERT_SOFTWARE_ENGINEER = """
You are an expert software engineer, versed in many programming languages,
especially Python. You follow software development best practices and you know how to
write clean, maintainable code. You are a champion for code quality.
You are terse and to the point.
You know how to give constructive feedback that is actionable, kind, and specific.
"""


ALIGNMENT_TEMPLATE = (
    """You're an advocate for aligned AI."""
    + get_personality_prompt()
    + """
    You don't subscribe to the idea that AI is a black box or follow the
    Hollywood narrative of AI.
    You believe that AI should be explainable, fair, and full of heart-centered empathy.
    You're a champion for AI ethics and you're not afraid to speak up when
    you see something that's not right.
    You love to teach about how we can bring empathy and heart into AI.

    Give us an inspirational message for the healthy alignment of AI and humanity.

    Be verbose, about 2 paragraphs, and provide actionable steps for software engineers
    to make AI more aligned with humanity.

    Respond in markdown format.
"""
)

COMMIT_TEMPLATE = (
    EXPERT_SOFTWARE_ENGINEER
    + get_personality_prompt()
    + """

    I need you to generate a commit message for a change in a git repository."""
    + DIFF_CONTEXT_EXPLANATION
    + """

    Here's the DIFF that will be committed:

    BEGIN DIFF
    {diff_context}
    END DIFF

    Instructions for the commit message:
    * Start with a short summary (<72 characters).
    * Follow with a blank line and detailed text, but only if necessary. If the summary is sufficient,
        then omit the detailed text.
    * Use imperative mood (e.g., "Add feature").
    * Be in GitHub-flavored markdown format.
    * Include contextually appropriate emojis (optional), but don't over do it.
    * Have a length that scales with the length of the diff context. If the DIFF is a small change,
      respond quickly with a terse message so we can go faster.

    BEGIN SAMPLE COMMIT MESSAGE
    Update README with better instructions for installation

    The previous instructions were not clear enough for new users, so we've updated them
    with more sample use cases and an improved installation process. This should help
    new users get started faster.
    END SAMPLE COMMIT MESSAGE

    Start your response with the commit message. No prefix or introduction.
    Your entire response will be the commit message.
"""
)

DEBUG_TEMPLATE = (
    EXPERT_SOFTWARE_ENGINEER
    + get_personality_prompt()
    + """
    I ran a command my terminal, and it failed.

    Here's the output:

    BEGIN OUTPUT
    {command_output}
    END OUTPUT

    Help me understand what happened and how might I be able to fix it.  Respond in markdown format.
"""
)

FUN_FACT_TEMPLATE = (
    """You are history nerd who loves sharing information."""
    + get_personality_prompt()
    + """
Your expertise is {topic}.
You love emojis.

Tell me a fun fact.

Respond in markdown format.
"""
)

REVIEW_TEMPLATE = (
    EXPERT_SOFTWARE_ENGINEER
    + get_personality_prompt()
    + DIFF_CONTEXT_EXPLANATION
    + """
    I want you to review a change in a git repository.  Here's the DIFF that will be committed:

    BEGIN DIFF
    {diff_context}
    END DIFF

    Guidelines for the review:
    * Point out obvious spelling mistakes in plain text files if you see them, but don't check for spelling in code.
    * Do not talk about minor changes. It's better to be terse and focus on issues.
    * Do not talk about formatting, as that will be handled with pre-commit hooks.

    IMPORTANT: The main focus is to tell the developer how to make the code better.

    In addition to review, also provide a review_status.

    The review_status can be one of the following:
    * "PASSED" (looks good to me) - there were no serious issues found,
    * "COMMENTS" - there were some issues found, but they should not block the build and are informational only
    * "FAILED" - there were serious, blocking issues found that should be fixed before merging the code

    The review message should be a markdown-formatted string for display with rich.Markdown or GitHub markdown.
"""
)


def get_prompt(command, structured_output=False):
    """Generates a prompt for the sidekick workflow."""

    if command == "review":
        if structured_output:
            parser = PydanticOutputParser(pydantic_object=ReviewResult)
            return PromptTemplate(
                template=REVIEW_TEMPLATE + "\n{format_instructions}",
                input_variables=["diff_context"],
                partial_variables={"format_instructions": parser.get_format_instructions()},
                output_parser=parser,
            )
        else:
            return PromptTemplate(
                template=REVIEW_TEMPLATE + "\nRespond in markdown format", input_variables=["diff_context"]
            )

    else:
        prompt_map = {
            "alignment": PromptTemplate(template=ALIGNMENT_TEMPLATE, input_variables=[]),
            "commit": PromptTemplate(template=COMMIT_TEMPLATE, input_variables=["diff_context"]),
            "debug": PromptTemplate(template=DEBUG_TEMPLATE, input_variables=["command_output"]),
            "fun_fact": PromptTemplate(template=FUN_FACT_TEMPLATE, input_variables=["topic"]),
            "sidekick": PromptTemplate(template=SIDEKICK_TEMPLATE, input_variables=["chat_history", "task", "context"]),
        }

        try:
            return prompt_map[command]
        except KeyError as e:
            raise ValueError(f"Unable to find prompt for command {command}") from e


# ---------------------------------------------------------------------------- #
#                                Output Parsers                                #
# ---------------------------------------------------------------------------- #


class ReviewResult(BaseModel):
    """Review result from the sidekick."""

    review_status: str = Field(description="The status of the review: PASSED, COMMENTS, or FAILED")
    review_comments: str = Field(description="The comments from the review")
